﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;

namespace Buckets
{

    public enum Methods
    {
        BFS,
        Volume,
        PathLength,
        SumVolume,
        Combine1,
        Combine2
    }

    public class Solver
    {
        /// <summary>
        /// Solves water jug problem.
        /// </summary>
        /// <param name="instance">Instance.</param>
        /// <param name="method"></param>
        /// <returns></returns>
        public ISolution Solve(Instance instance, Methods method)
        {
            IConfigCollection states;
            switch(method)
            {
                case Methods.BFS: states = new ConfigQueue();
                    break;
                case Methods.Volume:
                    states = new ConfigPriorityQueue(new VolumeComparer());
                    break;
                case Methods.PathLength:
                    states = new ConfigPriorityQueue(new PathLengthComparer());
                    break;
                case Methods.SumVolume:
                    states = new ConfigPriorityQueue(new SumVolumeComparer());
                    break;
                case Methods.Combine1:
                    states = new ConfigPriorityQueue(new CombineComparer(new SumVolumeComparer(), new PathLengthComparer()));
                    break;
                case Methods.Combine2:
                    states = new ConfigPriorityQueue(new CombineComparer(new VolumeComparer(), new SumVolumeComparer()));
                    break;
                default:
                    throw new ArgumentOutOfRangeException("Unsupported method");
            }

            states.Add(new Configuration(instance));

            while (states.Count > 0)
            {
                var config = states.Get();
                // solution found
                if (instance.IsSolution(config.Volumes))
                {
                    return new Solution()
                               {ID = instance.ID, NumOfNodes = states.NumOfStates - states.Count, Steps = config.Step};
                }
                // expand states
                states.Add(ExpandState(config));
            }

            // solution doesn't exists
            return null;
        }

        /// <summary>
        /// If isnt empty, empty it, otherwise return null.
        /// </summary>
        /// <param name="config"></param>
        /// <param name="idx"></param>
        /// <returns>Returns new Configuration or null if cant be created.</returns>
        public Configuration Empty(Configuration config, int idx)
        {
            Configuration newConfig = null;
            if (config.Volumes[idx] > 0)
            {
                newConfig = config.Clone();
                newConfig.CreateOperation = BucketOperations.Empty;
                newConfig.Volumes[idx] = 0;
                
            }

            return newConfig;
        }

        /// <summary>
        /// If is space in bucket to fill, fill it, otherwise return null.
        /// </summary>
        /// <param name="config"></param>
        /// <param name="idx"></param>
        /// <returns>Returns new Configuration or null if cant be created.</returns>
        public Configuration Fill(Configuration config, int idx)
        {
            Configuration newConfig = null;
            if (config.Volumes[idx] < config.Capacities[idx])
            {
                newConfig = config.Clone();
                newConfig.CreateOperation = BucketOperations.Fill;
                newConfig.Volumes[idx] = config.Capacities[idx];
            }

            return newConfig;
        }

        /// <summary>
        /// If is enought space in bucket1, pour bucket 2, otherwise return null.
        /// </summary>
        /// <param name="config"></param>
        /// <param name="idx1"></param>
        /// <param name="idx2"></param>
        /// <returns>Returns new Configuration or null if cant be created.</returns>
        public Configuration Pour(Configuration config, int fromIdx, int toIdx)
        {
            Configuration newConfig = null;
            if (config.Volumes[fromIdx] > 0 && config.Volumes[toIdx] < config.Capacities[toIdx])
            {
                newConfig = config.Clone();
                newConfig.CreateOperation = BucketOperations.Pour;
                byte volume = (byte) Math.Min(config.Volumes[fromIdx], config.Capacities[toIdx] - config.Volumes[toIdx]);
                newConfig.Volumes[toIdx] += volume;
                newConfig.Volumes[fromIdx] -= volume;
            }

            return newConfig;
        }

        public IEnumerable<Configuration> ExpandState(Configuration config)
        {
            var result = new LinkedList<Configuration>();

            for (int i = 0; i < config.Count; i++)
            {
                Configuration newConfig = null;
                if ((newConfig = Fill(config, i)) != null) result.AddLast(newConfig);
                if ((newConfig = Empty(config, i)) != null) result.AddLast(newConfig);

                for (int j = 0; j < config.Count; j++)
                {
                    if (i == j) continue;
                    if ((newConfig = Pour(config, i, j)) != null) result.AddLast(newConfig);
                }
            }
            return result;
        }

    }

    public class SumVolumeComparer : IComparer<Configuration>
    {

        public int Compare(Configuration x, Configuration y)
        {
            var valX = Sum(x.Volumes, x.Instance.Finals);
            var valY = Sum(y.Volumes, y.Instance.Finals);

            if (valX < valY) return 1;
            return -1;


        }

        private static int Sum(byte[] volumes, byte[] finals)
        {

            var sumDiff = Math.Abs(volumes.Sum(i => i) - finals.Sum(i => i));

            return sumDiff;
        }
    }

    public class VolumeComparer : IComparer<Configuration>
    {

        public int Compare(Configuration x, Configuration y)
        {
            var valX = RightVolumes(x.Volumes, x.Instance.Finals);
            var valY = RightVolumes(y.Volumes, y.Instance.Finals);

            if (valX > valY) return 1;
            return -1;
            
        }

        private static int RightVolumes(byte[] volumes, byte[] finals)
        {
            return volumes.Where((t, i) => t == finals[i]).Count();
        }
    }
    
    public class PathLengthComparer : IComparer<Configuration>
    {

        public int Compare(Configuration x, Configuration y)
        {
            var valX = x.Step;
            var valY = y.Step;

            if (valX < valY) return 1;
            return -1;

        }

    }

    public class CombineComparer : IComparer<Configuration>
    {
        private IComparer<Configuration> _comparer1;
        private IComparer<Configuration> _comparer2;

        public CombineComparer(IComparer<Configuration> comparer1, IComparer<Configuration> comparer2)
        {
            _comparer1 = comparer1;
            _comparer2 = comparer2;
        }

        public int Compare(Configuration x, Configuration y)
        {
            var res = _comparer1.Compare(x, y);
            if (res == 0) return _comparer2.Compare(x, y);
            return res;

        }

    }


}
